/*********************************************
 * vim: set sw=8 ts=8 si :
 * This program use a microcontroller as a (i2c
 * controlled) group of 8 off/on step relais. 
 * B0,....,B7 should be connected to buttons (contacts to ground).
 * D0,....,D7 are the outputs
 * C0,...C3 to dip switches (i2c address)
 * 1=connected to ground.
 * There is a software based delay to prevent spurious signals
 * from buttons.
 * Author: Renzo Davoli, Copyright: GPLv2
 * Based on avrm8ledtest.c
 * Author: Guido Socher, Copyright: GPL 
 * Chip type           : ATMEGA88
 * Clock frequency     : Internal clock 8 Mhz 
 *********************************************/
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/twi.h>
#include <inttypes.h>
#ifdef ATMEGA88
#define F_CPU 8000000UL  // 8 MHz
#else
#define F_CPU 1000000UL  // 1 MHz
#endif
#include <util/delay.h>


void delay_ms(unsigned int ms)
	/* delay for a minimum of <ms> */
{
	// we use a calibrated macro. This is more
	// accurate and not so much compiler dependent
	// as self made code.
	while(ms){
		_delay_ms(0.96);
		ms--;
	}
}


#define DMASK 0x07 /* 8 ms */
#define TWSR_STATUS_MASK 0xF8
#define TWCR_CMD_MASK 0x0F

unsigned char newB;
unsigned char oldB;
unsigned char diffB;
unsigned char status;
unsigned char unstable;
unsigned char step;
unsigned char value;
unsigned char d0,d1,d2,d3,d4,d5,d6,d7,d8;

static unsigned char i2cs_addr;

void i2cs_init(void)
{
	TWCR = 1<<TWINT | 1<<TWEA | 1<<TWEN | 1<<TWIE;
	TWAR = i2cs_addr;
}

void i2cs_command(unsigned char comm)
{
	switch (comm & 0xF8) {
		case 0x80:
			if (step)
				status &= ~(1<<(comm & 0x07));
			else
				PORTD &= ~(1<<(comm & 0x07));
			break;
		case 0x88:
			if (step)
				status |= (1<<(comm & 0x07));
			else
				PORTD |= (1<<(comm & 0x07));
			break;
	}
}


SIGNAL(SIG_2WIRE_SERIAL)
{
	unsigned char i2c_status= TWSR & TWSR_STATUS_MASK;
	switch (i2c_status) {
		/* START 0x60, 0x68 */
		case TW_SR_DATA_ACK:
			i2cs_command(TWDR);
			TWCR = (TWCR & TWCR_CMD_MASK) |  1<<TWINT | 1<<TWEA;
			break;
		case TW_ST_SLA_ACK:
			/* TX DATA 0xB8(fallthrough) */
		case TW_ST_DATA_ACK:
			TWDR = status;
			TWCR = (TWCR & TWCR_CMD_MASK) |  1<<TWINT | 1<<TWEA;
			break;
			/* TX DATA NAK or LAST 0xC0 0xC8 */
		case TW_NO_INFO:
			break;
		/*case TW_SR_SLA_ACK:
		case TW_SR_ARB_LOST_SLA_ACK:
		case TW_SR_DATA_NACK:
		case TW_SR_STOP:
		case TW_ST_DATA_NACK:
		case TW_ST_LAST_DATA:*/
		default:
			TWCR = (TWCR & TWCR_CMD_MASK) |  1<<TWINT | 1<<TWEA;
			break;
	}
}

int main(void)
{
#ifdef ATMEGA88
	// set the clock speed to 8MHz
	// set the clock prescaler. First write CLKPCE to enable setting of clock the
	// next four instructions.
	CLKPR=(1<<CLKPCE);
	CLKPR=0; // 8 MHZ
	delay_ms(10);
#endif

	/* INITIALIZE */
	DDRD = 0xFF; /* D ALL OUTPUTS */

	//DDRB = 0; /* B ALL INPUTS */
	PORTB = 0xFF;

	//DDRC = 0x00;
	PORTC = 0x0F;
	delay_ms(10);

	i2cs_addr= (0x20 | ((~PINC & 0x0F) << 1));
	step = (i2cs_addr & 0x10) == 0;
	i2cs_init();
	sei();

	while (1) {
		delay_ms(1);
		oldB=newB;
		newB=~PINB; /* B pins have pull up resistors */
		diffB=oldB ^ newB;
		unstable=0;
		if (diffB & 0x01) d0 = 1; else d0++; 
		if (diffB & 0x02) d1 = 1; else d1++;
		if (diffB & 0x04) d2 = 1; else d2++;
		if (diffB & 0x08) d3 = 1; else d3++;
		if (diffB & 0x10) d4 = 1; else d4++;
		if (diffB & 0x20) d5 = 1; else d5++;
		if (diffB & 0x40) d6 = 1; else d6++;
		if (diffB & 0x80) d7 = 1; else d7++;
		if(d0 & DMASK) unstable |= 0x01;
		if(d1 & DMASK) unstable |= 0x02;
		if(d2 & DMASK) unstable |= 0x04;
		if(d3 & DMASK) unstable |= 0x08;
		if(d4 & DMASK) unstable |= 0x10;
		if(d5 & DMASK) unstable |= 0x20;
		if(d6 & DMASK) unstable |= 0x40;
		if(d7 & DMASK) unstable |= 0x80;
		if (step) {
			status ^= (newB & ~value & ~unstable);
			value = (value & unstable) | (newB & ~unstable);
		} else
			status = (status & unstable) | (newB & ~unstable);

		if (step)
			PORTD=status;
	}
	return(0);
}
